#ifndef _BASIC_
#define _BASIC_ 1
#include <pthread.h>
#include <stdint.h>

/* Malloc with error checking */
void *malloc_check(size_t sz) ;
void *calloc_check(size_t nmemb, size_t size) ;

/* Error function */

void errexit(char *msg,int err) ;

/* Boxed int */
typedef struct {
  int v ;
} boxed_int_t ;

void free_boxed_int(boxed_int_t *p) ;
boxed_int_t *alloc_boxed_int(int v) ;

/* Threads */
void create_thread(pthread_t *th,void *(*f)(void *),void *x) ;
void create_thread_detached(pthread_t *th,void *(*f)(void *),void *x) ;
void *join_thread(pthread_t *th) ;

/* Mutex */
pthread_mutex_t *alloc_mutex(void) ;
void free_mutex(pthread_mutex_t *p) ;
void lock_mutex(pthread_mutex_t *p) ;
void unlock_mutex(pthread_mutex_t *p) ;

/* Condition variable */
pthread_cond_t *alloc_cond(void) ;
void free_cond(pthread_cond_t *p) ;

void wait_cond(pthread_cond_t *c, pthread_mutex_t *m) ;
void signal_cond(pthread_cond_t *c) ;
void broadcast_cond(pthread_cond_t *c) ;

/* Barrier */
pthread_barrier_t *alloc_barrier(unsigned count) ;
void free_barrier(pthread_barrier_t *p) ;
int wait_barrier(pthread_barrier_t *p) ;

/* Delay expressed in miliseconds */
void delay(unsigned int ms) ; 

/* Wall-clock timings, expressed in microsecs */

typedef uint64_t tsc_t ;
#define PTSC PRIu64

/* Result in micro-seconds */
tsc_t timeofday(void) ;
double tsc_millions(tsc_t t) ;

/* Atomic builtins */
#define DECR(x) __sync_add_and_fetch(x,-1)
#define INCR(x) __sync_fetch_and_add(x,1)
#define GET(x)  __sync_fetch_and_add(x,0)
#define SYNC __sync_synchronize
#endif
