#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/time.h>
#include <time.h>
#include <stdatomic.h>
#include <threads.h>


/*************/
/* Utilities */
/*************/

static int verbose = 0 ;

void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [size] [nproc] [capacity]\n",pgm) ;
}

void error_exit(char *msg) {
  fprintf(stderr,"%s\n",msg) ;
  exit(2) ;
}

void *malloc_check(size_t sz) {
  if (sz == 0) return NULL ;
  void *p = malloc(sz) ;
  if (!p) {
    perror("malloc") ;
    exit(2) ;
  }
  return p ;
}

void delay(unsigned int ms) {
  struct timespec req,rem ;
  unsigned int sec = ms / 1000 ;
  long ns = (ms % 1000) * 100000L ;
  req.tv_sec = sec ; req.tv_nsec = ns ;
  while (nanosleep(&req,&rem)) {
    if (errno == EINTR) {
      req = rem ;
    } else {
      error_exit("nanosleep") ;
    }
  }
}

/***************************/
/* Synchronisation barrier */
/***************************/

// Notice Lock Free coding, do not imitate!

typedef struct {
  unsigned int nprocs ;
  atomic_uint count ;
  atomic_int sense ;
} barrier_t ; 

barrier_t *alloc_barrier(unsigned int nprocs) {
  barrier_t *p = malloc_check(sizeof(*p)) ;
  p->nprocs = p->count = nprocs ;
  p->sense = 0 ;
  return p ;
}

void free_barrier(barrier_t *p) {
  free(p) ;
}

void wait_barrier(barrier_t *p) {
  int sense = atomic_load(&p->sense) ;
  unsigned int rem = atomic_fetch_sub(&p->count,1)-1;
  if (rem == 0) {
    atomic_store(&p->count,p->nprocs) ;
    atomic_store(&p->sense,!sense) ;
  } else {
    while (atomic_load(&p->sense) == sense) ;
  }
}


/**************/
/* Semaphores */
/**************/

/* Semaphore proper */

typedef struct {
  volatile int nfree ;
  mtx_t *mutex ;
  cnd_t *cond ;
} semaphore_t ;


// TODO
semaphore_t *alloc_semaphore(int capacity) {
  return NULL ;
}

// TODO
void free_semaphore(semaphore_t *p) {

}

// TODO
void wait_semaphore(semaphore_t *p) {
}

// TODO
void post_semaphore(semaphore_t *p) {
}

/***********/
/* Threads */
/***********/

typedef struct {
  int size ;
  barrier_t *b ;
  semaphore_t *sem ;
} common_t ;

typedef struct {
  int id ;
  common_t *common ;
} ctx_t ;


int T1(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;
  for (int k = q->size-1 ; k >= 0 ; k--) {
    wait_semaphore(q->sem) ;
    printf("+") ;
    delay(50) ;
    printf("-") ;
    post_semaphore(q->sem) ;
    wait_barrier(q->b) ;
    if (p->id == 0) printf("\n") ;
    wait_barrier(q->b) ;
  }
  return 0 ;
}

void run(int size, int nprocs, int m) {
  common_t c ;
  c.b = alloc_barrier(nprocs) ;
  c.sem = alloc_semaphore(m) ;
  c.size = size ;

  thrd_t th[nprocs] ;
  ctx_t a[nprocs] ;
  for (int k = 0 ; k < nprocs ; k++) {
    // TODO create thread, with approprite arguments
  }
  for (int k = 0 ; k < nprocs ; k++) {
    // TODO join threads
  }
  free_semaphore(c.sem) ;
  free_barrier(c.b) ;
}

int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  int size = 10 ;
  int n = 4 ;
  int m = 2 ;
  ++argv ;
  --argc ;
  while (argc > 0) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    ++argv;
    --argc ;
  }
  if (argc == 1) {    
    size = atoi(argv[0]) ;
  } else if (argc == 2) {
    size = atoi(argv[0]) ;
    n = atoi(argv[1]) ;
  } else if (argc > 2) {
    size = atoi(argv[0]) ;
    n = atoi(argv[1]) ;
    m = atoi(argv[2]) ;
  }
  run(size,n,m) ;
  return 0 ;
}
